#include "cmdparser.hpp"
#include <cstring>

namespace fs = boost::filesystem;

using namespace std;

/* This class holds the complete configuration for mbox2eml. */
Configuration::Configuration() :
    input_file(),
    output_dir(),
    unbitstuff(false),
    unbitstuff_char('>'),
    unbitstuff_targets(">"),
    trailing_space(true)
    {}

/* constructs the command line parser */
Cmdline_parser::Cmdline_parser() {}

/* Here, the initialization of popt and all the parsing is done.*/
Configuration*
Cmdline_parser::parse(const int argc, const char* argv[]) {
    // All available options.
    static const struct poptOption options[] =
	  {
	    {"unbitstuff-char",    'c', POPT_ARG_STRING, NULL, 'c'},
	    {"output-dir",         'd', POPT_ARG_STRING, NULL, 'd'},
	    {"input-file",         'f', POPT_ARG_STRING, NULL, 'f'},
	    {"help",               'h', POPT_ARG_NONE,   NULL, 'h'},
	    {"notrailing-space",   's', POPT_ARG_NONE,   NULL, 's'},
	    {"unbitstuff-targets", 't', POPT_ARG_STRING, NULL, 't'},
	    {"unbitstuff",         'u', POPT_ARG_NONE,   NULL, 'u'},
	    {NULL, 0, 0, NULL, 0}
	  };

    // The -thunderbird shortcut.
    const char** alias_tb_argv = new const char*[6];
    alias_tb_argv[0] = "-u";
    alias_tb_argv[1] = "-s";
    alias_tb_argv[2] = "-c";
    alias_tb_argv[3] = " ";
    alias_tb_argv[4] = "-t";
    alias_tb_argv[5] = " >";

    static const struct poptAlias alias_tb =
        {"thunderbird", '\0', 6, alias_tb_argv};
    // Get the context.
    poptContext context = poptGetContext("mbox2eml", argc, argv, options, 0);
    // Add the -thunderbird shortcut.
    int c = poptAddAlias(context, alias_tb, 0);

    Configuration* config = new Configuration();

    // parse the command line options
    while((c = poptGetNextOpt(context)) > 0) {
        const char *optarg = poptGetOptArg(context);
        switch(c) {
            case 'c':
                if(strlen(optarg) != 1) {
                    cerr << "-c: argument must be exactly ONE character!" << endl;
                    exit(1);
                }
                config->unbitstuff_char = optarg[0];
                #ifdef CMDDEBUG
                cout << "option: " << (char)c << " argument: " << optarg << endl;
                #endif //CMDDEBUG
                break;
            case 'd':
                config->output_dir = check_dir(optarg);
                #ifdef CMDDEBUG
                cout << "option: " << (char)c << " argument: " << optarg << endl;
                #endif //CMDDEBUG
                break;
            case 'f':
                config->input_file = check_file(optarg);
                #ifdef CMDDEBUG
                cout << "option: " << (char)c << " argument: " << optarg << endl;
                #endif //CMDDEBUG
                break;
            case 'h':
                #ifdef CMDDEBUG
                cout << "option: " << (char)c << endl;
                #endif //CMDDEBUG
                cout << usage() << endl;
                exit(0);
            case 's':
                config->trailing_space = false;
                #ifdef CMDDEBUG
                cout << "option: " << (char)c << endl;
                #endif //CMDDEBUG
                break;
            case 't':
                config->unbitstuff_targets = optarg;
                #ifdef CMDDEBUG
                cout << "option: " << (char)c << " argument: " << optarg << endl;
                #endif //CMDDEBUG
                break;
            case 'u':
                config->unbitstuff = true;
                #ifdef CMDDEBUG
                cout << "option: " << (char)c << endl;
                #endif //CMDDEBUG
                break;
            default:
                cerr << "unknown option: " << (char)c << endl;
                exit(1);
        }
    }

    // error handling
    if(c < -1) {
        cerr << poptBadOption(context, 0) << ": " << poptStrerror(c) << endl;
    }

    // Leftover Arguments
    if(poptPeekArg(context) != NULL) {
        cerr << "No arguments without option switch allowed:" << endl;
        const char *xtra_arg;
        while((xtra_arg = poptGetArg(context)) != NULL) {
            cerr << xtra_arg << endl;
        }
    }

    poptFreeContext(context);

    // assure that all obligatory Options are set correctly
    if(!check_necessary_options_set(config)) {
        cerr << usage(true);
        exit(1);
    }

    return config;
}

/* Here, all obligatory options are checked for existance. */
bool
Cmdline_parser::check_necessary_options_set(Configuration* config) {
    bool result = true;

    if(config->input_file.empty()) {
        cerr << "You must specify an input-file!" << endl;
        result = false;
    }

    if(config->output_dir.empty()) {
        cerr << "You must specify an output-directory!" << endl;
        result = false;
    }

    return result;
}

/* This function checks the submitted directory name for:
 * - correctness of the name
 * - existance
 * If the directory does not exist and the name is valid,
 * a new dirctory is created.
 *
 * If the directory name exists but is no directory, an error is raised.
 *
 * Returns: the verified path of the directory
 */
fs::path
Cmdline_parser::check_dir(const char* dir_str) {
    // compose the output directory's path
    fs::path full_path;
    try {
        fs::path dir_path(dir_str);
        full_path = fs::system_complete(dir_path);
    }
    catch(...) {
        cerr << "Directory name is not valid: " << dir_str << endl;
        exit(1);
    }
    // create the output directory if necessary
    bool exists = false;
    try {
        exists = fs::exists(full_path);
    }
    catch(const fs::filesystem_error& ex) {
        cerr << "Could not check if dir exists: " << ex.what() << endl;
        exit(1);
    }
    if(!exists) {
        try {
            exists = fs::create_directory(full_path);
        }
        catch(...) {}
        if(!exists) {
            cerr << "Could not create directory: " << full_path << endl;
            exit(1);
        }
    }
    // does the ouput path represent a directory?
    if(!fs::is_directory(full_path)) {
        cerr << "This is a file and not a directory: " << full_path << endl;
        exit(1);
    }

    return full_path;
}

/* This function checks the submitted file name for existance.
 * If the file name exists but is no file, an error is raised.
 *
 * Returns: the verified path of the file
 */
fs::path
Cmdline_parser::check_file(const char* file_str) {
    // compose the input file's path
    fs::path full_path;
    try {
        fs::path file_path(file_str);
        full_path = fs::system_complete(file_path);
    }
    catch(...) {
        cerr << "File name is not valid: " << file_str << endl;
        exit(1);
    }
    // does the input file exist?
    try {
        if (!fs::exists(full_path)) {
            cerr << "File does not exist: " << full_path << endl;
            exit(1);
        }
    }
    catch(const fs::filesystem_error& ex) {
        cerr << "Could not check if file exists: " << ex.what() << endl;
        exit(1);
    }
    // is it really a file?
    if(fs::is_directory(full_path)) {
        cerr << "This is a directory and not a file: " << full_path << endl;
        exit(1);
    }

    return full_path;
}

/* returns the usage instructions
 *
 * if oneliner==true only one line is returned
 */
string
Cmdline_parser::usage(bool oneliner) {
    ostringstream os;

    os << "Usage: mbox2eml [options] -f MBOX-FILE -d DEST-DIR\n";
    if(oneliner) {
        return os.str();
    }

    os << "Obligatory Options:\n"
	   << "  --output-dir | -d :\n"
	   << "    This is the output directory, where the eml-files are being saved.\n"
	   << "  --input-file | -f :\n"
	   << "    This is the input file which shall be split into eml-files.\n"
	   << "    This file MUST be in mbox format as described here:\n"
	   << "    http://www.die.net/doc/linux/man/man5/mbox.5.html\n"
	   << "Optional Options:\n"
	   << "  --help | -h :\n"
	   << "    Displays this text.\n"
	   << "Advanced Options:\n"
	   << "THESE OPTIONS ARE VERY DANGEROUS!!!\n"
	   << "PLEASE READ THE README-FILE VERY CAREFULLY.\n"
	   << "  --unbitstuff | -u :\n"
	   << "    Turns on the 'unbitstuff' functionality.\n"
	   << "  --unbitstuff-char | -c :\n"
	   << "    This is the character, which will be removed when 'unbitstuff'ing.\n"
	   << "    Default: '>'\n"
       << "  --notrailing-space | -s :\n"
	   << "    This option triggers if a trailing space after 'From' is necessary.\n"
	   << "    Default: trailing space is necessary\n"
	   << "  --unbitstuff-targets | -t :\n"
	   << "    These are the characters which may be between the 'unbitstuff-char'\n"
	   << "    and 'From'. If only these characters are found, the affected line is\n"
	   << "    'unbitstuff'ed.\n"
	   << "    Default: \">\"\n"
	   << "  --thunderbird:\n"
	   << "    This is an alias for -u -s -c \" \" -t \" >\"\n"
	   << "    This option enables the 'unbitstuff'ing for mbox-files produced by\n"
	   << "    Mozilla Thunderbird.\n"
	   << "    Thundebird violates the common mbox practice to bitstuff every line\n"
	   << "    starting with '[>]*From ' with a single '>'. Instead of this,\n"
	   << "    Thunderbird bitstuffs every line starting with '[ >]*From' with a\n"
	   << "    whitespace.\n"
	   << "    Using this option may corrupt digitally signed (i.e. S/MIME or PGP/GPG)\n"
	   << "    mails that have been encrypted using Thunderbird or EnigMail as\n"
	   << "    Thunderbird is internally working with the 'false' bitstuffed\n"
	   << "    mbox-format. So, if you try to fix this wrong format, you will modify\n"
	   << "    the content of the mail but the signature will not reflect the\n"
	   << "    modifications. And so the signature will be invalid.\n";

    return os.str();
}

