#ifndef MBOXPARSER_HPP
#define MBOXPARSER_HPP 1

#include "assure.hpp"
#include "cmdparser.hpp"

#include "boost/algorithm/string.hpp"
#include "boost/algorithm/string/predicate.hpp"
#include "boost/filesystem/operations.hpp"
#include "boost/filesystem/path.hpp"

#include <deque>
#include <iomanip>
#include <sstream>
#include <string>


/* #define MBOXPARSERDEBUG */

struct Mail {
    public:
        /* contains the mail header line by line */
        std::deque <std::string> header;
        /* contains the mail body line by line */
        std::deque <std::string> body;

        /* Initializes the new object representing a mail.
         * A mail consists of a header and a body. The header and the body are parsed
         * and initialized implicitly by this constructor.
         *
         * The deque must contain the mail line by line.
         */
        Mail(std::deque <std::string> *mail, Configuration *config);

        /* Destructor: Clears the header and body of this mail. */
        ~Mail();

        /* Writes this mail to an eml file.
         * The destination file name is computed depending on the subject of this mail.
         *
         * The outdpath represents the output directory.
         * The outf_counter is used for consecutive numbers for mails without subject.
         */
        void store(const boost::filesystem::path outdname, long* outf_counter);

    private:
        /* Splits the submitted deque (containing the mail line by line) into two
         * deques, the header and the body.
         * Unbitstuffing is performed implicitly depending on the Configuration.
         */
        void parseHeader(std::deque <std::string> *mail, Configuration *config);

        /* Parses a header field from this mail.
         *
         * field must be i.e. "subject:"
         */
        std::string parseHeaderField(std::string field);

        /* Generates unique file names by appending consecutive numbers to the end of
         * the file names if necessary.
         * It is ensured that the file names won't exceed 31 Characters. So, the
         * file names should be quite portable.
         */
        std::string generateUniqueFilename(std::string subject, const boost::filesystem::path outdpath, long* outf_counter);

        /* Replaces all characters that are not [0-9a-zA-Z-.] by an underscore and
         * replaces multiple consecutive underscores by a single underscore.
         * This should lead to maximum portability of filenames.
         */
        void fixFilename(std::string *filename);

        /* Reverts the bitstuffing which mbox-writers perform in order to ensure that
         * only one line per mail starts with 'From '.
         */
        void unbitstuff(std::string *textline, Configuration *config);

        /* converts numbers to strings */
        template <class TYPE> std::string NumberToString(const TYPE & t, const int fillup);
};

#endif /* !MBOXPARSER_HPP */

